;; photoblogger
;; Copyright (C) 2004 Andy Wingo

;; This program is free software; you can redistribute it and/or    
;; modify it under the terms of the GNU General Public License as   
;; published by the Free Software Foundation; either version 2 of   
;; the License, or (at your option) any later version.              
;;                                                                  
;; This program is distributed in the hope that it will be useful,  
;; but WITHOUT ANY WARRANTY; without even the implied warranty of   
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
;; GNU General Public License for more details.                     
;;                                                                  
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, contact:
;;
;; Free Software Foundation           Voice:  +1-617-542-5942
;; 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
;; Boston, MA  02111-1307,  USA       gnu@gnu.org

;;; Commentary:
;;
;;; Code:

(define-module (photoblogger)
  :use-module (srfi srfi-1)
  :use-module (srfi srfi-13)
  :use-module (sxml simple)
  :use-module (gnome gtk)
  :use-module (gnome glade)
  :use-module (gnome vfs)
  :use-module (gnome gnome-ui)
  :use-module (photoblogger config)
  :export (<photo> photoblogger-init
           glade-xml main-window tree-view status
           orig-pixbuf source-uri base-name pformat description))

(define glade-xml (glade-xml-new glade-file))
(define main-window (get-widget glade-xml "window"))
(define tree-view (get-widget glade-xml "treeview"))
(define status (get-widget glade-xml "status"))

(define (resize-factor p x y)
  (let* ((orig-x (get-width p))
         (orig-y (get-height p)))
    (cond
     ((and (< orig-x x) (< orig-y y))
      1.0)
     ((> (/ orig-x orig-y) (/ x y))
      (/ x orig-x))
     (else
      (/ y orig-y)))))

(define (resize-pixbuf p x y)
  (let* ((orig-x (get-width p))
         (orig-y (get-height p))
         (factor (resize-factor p x y))
         (new-x (inexact->exact (* orig-x factor)))
         (new-y (inexact->exact (* orig-y factor))))
    (cond
     ((= factor 1.0) p)
     (else
      (let ((new (gdk-pixbuf-new 'rgb #f 8 new-x new-y)))
        (if (get p 'has-alpha)
            (set! new (add-alpha new #f #\x #\x #\x)))
        (gdk-pixbuf-scale
         p new 0 0 new-x new-y 0 0 factor factor 'bilinear)
        new)))))

;; would be nice to have symbol-macrolet
(define-macro (with-slots obj slot-list . body)
  `(if ,obj
       (let ,(map (lambda (s) `(,s (slot-ref ,obj ',s))) slot-list)
         ,@body)))

;; the 'pixbuf gobject property is the thumbnail
(define-class <photo> (<gtk-image>)
  (orig-pixbuf :init-keyword :orig-pixbuf :accessor orig-pixbuf)
  (source-uri :init-keyword :source-uri :accessor source-uri)
  (base-name :init-keyword :base-name :accessor base-name)
  (pformat :init-keyword :pformat :accessor pformat)
  (description :init-keyword :description :accessor description :init-value #f)
  (already-uploaded :init-keyword :already-uploaded :init-value #f))

(define-method (initialize (obj <photo>) initargs)
  (next-method)
  (if (not (and-map (lambda (x) (slot-bound? obj x))
                    '(pformat base-name)))
      (let ((s (gnome-vfs-uri-extract-short-name
                (gnome-vfs-uri-new (source-uri obj)))))
        (set! (pformat obj)
              (if (string=? (gnome-vfs-get-mime-type (source-uri obj))
                            "image/jpeg")
                  "jpeg" "png"))
        (set! (base-name obj)
              (substring s 0 (or (string-rindex s #\.)
                                 (string-length s))))))
  (set! (orig-pixbuf obj)
        (or (gnome-gdk-pixbuf-new-from-uri (source-uri obj))
            (throw 'image-load-failed)))
  (set obj 'pixbuf (resize-pixbuf (orig-pixbuf obj) 64 64)))

(use-modules (photoblogger preferences)
             (photoblogger dialogs)
             (photoblogger state))

(define (status-set! status-string)
  (set status 'label status-string))

(define-macro (with-status status-string . body)
  (let ((old-str (gensym))
        (str (gensym)))
    `(let ((,old-str (get-text ,status))
           (,str ,status-string))
       (dynamic-wind
           (lambda () (status-set! ,str))
           (lambda () ,@body)
           (lambda () (status-set! ,old-str))))))

(define (add-photo-from-uri str)
  (and=> (false-if-exception (make <photo> :source-uri str))
         append-photo))

(define (%append-photo photo)
  (let* ((model (get-model tree-view))
         (iter (append model)))
    (set-value model iter 0 photo)))

(define (append-photo photo)
  (%append-photo photo)
  (select-last-photo)
  (state-set! 'photos
              (cons photo (state-ref 'photos)))
  photo)

(define (get-current-photo)
  (call-with-values (lambda () (get-selected (get-selection tree-view)))
    (lambda (model iter)
      (and iter (get-value model iter 0)))))

(define (select-last-photo)
  (let ((len (iter-n-children (get-model tree-view) #f)))
    (if (positive? len)
        (select-path (get-selection tree-view) (list (1- len))))))

(define (remove-selected-photo)
  (call-with-values (lambda () (get-selected (get-selection tree-view)))
    (lambda (model iter)
      (if iter
          (begin
            (state-set! 'photos
                        (delete (get-value model iter 0)
                                (state-ref 'photos)))
            (remove model iter)
            (select-last-photo))))))

(define (get-clipboard)
  (gtk-clipboard-get-for-display
   (get-display main-window)
   (gdk-atom-intern "CLIPBOARD")))

(define (copy-selected-photo)
  (call-with-values (lambda () (get-selected (get-selection tree-view)))
    (lambda (model iter)
      (and iter
           (let ((photo (get-value model iter 0)))
             (with-slots photo (description base-name pformat)
               (let ((uri (build-photo-uri 
                           (state-ref 'http-url-base)
                           photo)))
                 (gtk-clipboard-set-text
                  (get-clipboard)
                  (with-output-to-string
                    (lambda () (sxml->xml
                                `(img
                                  (@ (src ,uri)
                                     (alt ,description))))))))))))))

(define prompt-for-file-name
  (let ((w #f))
    (lambda ()
      (if (not w)
          (set! w (make <gtk-file-chooser-dialog>
                    :action 'open :title "Add Photo..."
                    :destroy-with-parent #t)))
      (set-transient-for w main-window)
      (for-each
       (lambda (id)
         (add-button w (gtk-stock-id id)
                     (genum->value (make <gtk-response-type> :value id))))
       '(cancel ok))
      (show w)
      (let* ((response (genum->symbol
                        (make <gtk-response-type> :value (run w))))
             (filename (get-uri w)))
        (hide w)
        (case response
          ((ok) filename)
          (else #f))))))

(define (setup-treeview)
  ;; blow away any old state so we can reload the func
  (map (lambda (c) (remove-column tree-view c)) (get-columns tree-view))
  (let ((model (gtk-list-store-new (list <photo>)))
        (selection (get-selection tree-view))
        (p-col (make <gtk-tree-view-column>))
        (p-col-renderer (make <gtk-cell-renderer-pixbuf>))
        (name-col (make <gtk-tree-view-column>))
        (name-col-renderer (make <gtk-cell-renderer-text>)))
    (set-model tree-view model)
    (for-each %append-photo (state-ref 'photos))
    (pack-start p-col p-col-renderer #f)
    (append-column tree-view p-col)
    (set-cell-data-func
     p-col p-col-renderer
     (lambda (col cel model iter)
       (let ((photo (get-value model iter 0)))
         (and photo (set cel 'pixbuf (get photo 'pixbuf))))))
    (pack-start name-col name-col-renderer #t)
    (append-column tree-view name-col)
    (set-cell-data-func
     name-col name-col-renderer
     (lambda (col cel model iter)
       (with-slots (get-value model iter 0) (base-name already-uploaded)
         (set cel 'editable (not already-uploaded))
         (set cel 'markup (format #f "<big><b>~A</b></big>~A" base-name
                                  (if already-uploaded "\n(uploaded)" ""))))))
    (connect name-col-renderer 'edited
             (lambda (col path newtext)
               (let ((photo (get-value model (get-iter model path) 0)))
                 (set! (base-name photo) newtext)
                 (state-set! 'photos (state-ref 'photos))
                 (emit selection 'changed))))
    (connect
     selection 'changed
     (lambda (selection)
       (call-with-values (lambda () (get-selected selection))
         (lambda (model iter)
           (if iter
               (let ((photo (get-value model iter 0)))
                 (with-slots photo (base-name orig-pixbuf)
                   (let* ((orig-x (get-width orig-pixbuf))
                          (orig-y (get-height orig-pixbuf))
                          (factor (resize-factor
                                   orig-pixbuf
                                   (state-ref 'scaled-image-width)
                                   (state-ref 'scaled-image-height)))
                          (new-x (inexact->exact (* factor orig-x)))
                          (new-y (inexact->exact (* factor orig-y))))
                     (status-set!
                      (format #f "<b>~A</b>: ~Ax~A -> ~Ax~A"
                              base-name orig-x orig-y new-x new-y)))))
               (status-set! "Drag an image onto the window to add."))))))
    (emit selection 'changed)))

(define (build-uri dirname basename)
  (or (gnome-vfs-uri-make-full-from-relative
       (gnome-vfs-make-uri-from-input dirname)
       basename)
      (throw 'gnome-vfs-error 'error-invalid-uri)))

(define (pformat->extension pformat)
  (if (string=? pformat "jpeg") ".jpg" ".png"))

(define (build-photo-uri dirname p)
  (with-slots p (base-name pformat)
    (build-uri dirname
               (string-append base-name (pformat->extension pformat)))))

(define (replace-file? p)
  (let ((base-name (slot-ref p 'base-name)))
    (eq? (question
          (format #f "The file \"~A\" already exists on the remote server. ~A"
                  base-name "Would you like to replace it?")
          "If you replace an existing file, its contents will be overwritten."
          '(no "Skip" #f)
          '(yes "Replace" #f))
         'yes)))

(define (make-output-port p)
  (let ((uri (build-photo-uri (state-ref 'write-url-base) p)))
    (define (open exclusive?)
      (gnome-vfs-create uri 'write exclusive? #o644))
    (catch 'gnome-vfs-error
           (lambda ()
             (open #t))
           (lambda (key type)
             (if (and (eq? type 'error-file-exists)
                      (replace-file? p))
                 (open #f)
                 (throw key type))))))

(define (list-index* l k eq?)
  (let loop ((n 0) (l l))
    (and (not (null? l))
         (if (eq? (car l) k) n (loop (+ n 1) (cdr l))))))

(define (set-photo-uploaded! p)
  (slot-set! p 'already-uploaded #t)
  (state-set! 'photos (state-ref 'photos))
  (let* ((model (get-model tree-view))
         (path (list (list-index* (reverse (state-ref 'photos)) p equal?)))
         (iter (get-iter model path)))
    (row-changed model path iter)))

(define (upload-files)
  (for-each
   (lambda (p)
     (catch 'gnome-vfs-error
            (lambda ()
              (let ((port (make-output-port p)))
                (save-to-port 
                 (resize-pixbuf (slot-ref p 'orig-pixbuf)
                                (state-ref 'scaled-image-width)
                                (state-ref 'scaled-image-height))
                 port
                 (pformat p)
                 '())
                (close-port port)
                (set-photo-uploaded! p)))
            (lambda (key type)
              (error-message
               (format #f "The file \"~A\" could not be uploaded."
                       (slot-ref p 'base-name))
               (gnome-vfs-result-to-string type)))))
   (filter
    (lambda (x) (not (slot-ref x 'already-uploaded)))
    (state-ref 'photos))))

;; signal handlers
(define (on-add-photo . args)
  (and=> (prompt-for-file-name)
         add-photo-from-uri))

(define (on-remove-photo . args)
  (remove-selected-photo))

(define (on-copy-photo . args)
  (copy-selected-photo))

(define (on-drag-data-received w drag-context x y selection-data info t)
  (let* ((data (gtk-selection-data-get-as-string selection-data))
         (uri-string (substring data 0 (1- (string-index data #\newline))))
         (path (gnome-vfs-get-local-path-from-uri uri-string)))
    (and=> path add-photo-from-uri)
    (gtk-drag-finish drag-context #t #t)))
   
(define (on-preferences . args)
  (show-preferences))

(define (on-delete-window . args)
  (exit 0))

(define (on-upload . args)
  (upload-files))

(define (animate-ui)
  (setup-treeview)
  (gtk-drag-dest-set main-window 'all '("text/plain") 'copy)
  (connect main-window 'delete-event on-delete-window)
  (connect main-window 'drag-data-received on-drag-data-received)
  (connect (get-widget glade-xml "add_button") 'clicked on-add-photo)
  (connect (get-widget glade-xml "remove_button") 'clicked on-remove-photo)
  (connect (get-widget glade-xml "copy_button") 'clicked on-copy-photo)
  (connect (get-widget glade-xml "preferences_button") 'clicked on-preferences)
  (connect (get-widget glade-xml "upload_button") 'clicked on-upload))

(define (photoblogger-init)
  (gnome-authentication-manager-init) ;; needed for keyring interaction
  (animate-ui)
  (show main-window)
  (if (string-null? (state-ref 'write-url-base))
      (show-preferences)))
